/* $Header: main.cv  1.6  91/06/07 10:53:56  Exp $ */
/* Main Function */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "reg.h"

extern int OutputNfaTree(nfa_state_ptr);
extern int OutputDfaTree(dfa_state_ptr);

char used[CHAR_STAFF];

void FatalError(char *, int);

/* DFAが文字列を受理するかどうか調べる */
int Test(dfa_state_ptr dfa, char *str)
{
	for ( ; *str != '\0'; str++) /* 一文字ずつ順に調べる */
		if (dfa->transition[0x00ff & (int)(*str)] == NULL)
			 /* 遷移先がなければ FALSEを返す */
			return (FALSE);
		else
			/* あれば、次の状態に進む */
			dfa = dfa->transition[0x00ff & (int)(*str)];

	if (dfa->condition == GOAL)
		return (TRUE);
	else
		return (FALSE);
}


/* 致命的エラーの処理 */
void FatalError(char *s, int n)
{
	printf("\nfatal error : %s (%d)\n", s, n);
	exit (-1);
}


#if DEBUG >= 1
nfa_state_ptr f[100];
int fp=0;

void PrintNfaState(nfa_state_ptr n)
{
	int i;

	for (i=0; i<fp; i++) {
		if (f[i] == n)
			return;
	}
	f[fp] = n;
	fp++;
	if (n->input[0] != EMPTY) {
	printf("[%0.4x]\t",n);
		switch (n->input[0]) {
		 case GOAL :	printf("in[0]:GOAL");
		 		break;
		 case EPSILON : printf("in[0]:EPSILON");
		 		break;
		 default :
				printf("in[0]:%c\t",n->input[0]);
		}
		printf("\ttransition[0]:%0.4x\n", n->transition[0]);
		if (n->transition[0] != NULL)
			PrintNfaState(n->transition[0]);
	}
	if (n->input[1] != EMPTY) {
	printf("[%0.4x]\t",n);
		switch (n->input[1]) {
		 case EPSILON : printf("in[1]:EPSILON");
		 		break;
		 default :
				printf("in[1]:%c\t",n->input[1]);
		}
		printf("\ttransition[1]:%0.4x\n", n->transition[1]);
		if (n->transition[0] != NULL)
			PrintNfaState(n->transition[1]);
	}
}
#endif


#if DEBUG == 2
void PrintDfaState(dfa_state_ptr d)
{
	int i;
	printf("\n");
	for (i=(int)' '; i<CHAR_STAFF; i++)
		printf("%c",(used[i]?used[i]:'@'));
	printf("\n");

	for ( ; d != NULL; d = d->next) {
		printf("(%0.4x) closure:%0.4x next:%0.4x %s\n",
		d, d->set_of_state, d->next,
		(d->condition ? "FINAL" : ""));
		PrintClosure(d->set_of_state);
	}
}

void PrintClosure(closure_ptr c)
closure_ptr c;
{
	for ( ; c != NULL; c = c->next) {
		printf("  %0.4x", c->state);
	}
	printf("\n");
}
#endif

int main(int argc, char *argv[])
{
	nfa_state_ptr reg2nfa(), start_of_nfa;
	dfa_state_ptr nfa2dfa(), start_of_dfa;

	if (argc != 3) {
		printf("argument count is wrong!!\n");
		printf("usage: reg <regular expression> <pattern>\n");
		exit (1);
	}

	bzero(used, CHAR_STAFF);

	start_of_nfa = reg2nfa(argv[1]);  /* 正規表現からNFAを作成 */
#if DEBUG >= 1
	PrintNfaState(start_of_nfa);
#endif
	printf("===== NFA =====\n");      /* 作成したNFAを表示 */
	printf("%d\n", OutputNfaTree(start_of_nfa));
	printf("\n\n");

	start_of_dfa = nfa2dfa(start_of_nfa); /* NFAをもとにDFAを作成 */
#if DEBUG == 2
	PrintDfaState(start_of_dfa);
#endif
	printf("===== DFA =====\n");      /* DFAを表示 */
	printf("%d\n", OutputDfaTree(start_of_dfa));
	printf("\n\n");

	if (Test(start_of_dfa, argv[2]))  /* 文字列がDFAに受理されるか? */
		printf("Yes\n");
	else
		printf("No\n");
	
	return (0);
}

