/* $Header: nfa2dfa.cv  1.6  91/06/07 10:53:58  Exp $ */
/* from Nondeterministic Finite Automata to Deterministic Finite Automata */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "reg.h"

extern char used[CHAR_STAFF];

static dfa_state_ptr start_of_dfa;
static closure_ptr cp;

void FatalError(char *, int);

int EqualClosure(closure_ptr, closure_ptr);
void MarkFinalState(dfa_state_ptr);
void EpsilonClosure(nfa_state_ptr);
int AddMember(nfa_state_ptr);
void MakeDfa();
void AddDfaState(dfa_state_ptr);
void FreeClosure(closure_ptr);

/* NFAからDFAを作る */
dfa_state_ptr nfa2dfa(nfa_state_ptr nfa)
{
	dfa_state_ptr MakeNewDfaState();

	cp = NULL;
	(void)AddMember(nfa);
	EpsilonClosure(nfa);
	start_of_dfa = MakeNewDfaState();
	start_of_dfa->set_of_state = cp;

	MakeDfa();
	MarkFinalState(start_of_dfa);
	return (start_of_dfa);
}


/* ε-CLOSUREを作る */
void EpsilonClosure(nfa_state_ptr nfa)
{
	if (nfa->input[0] == EPSILON)
		if (AddMember(nfa->transition[0]))
			EpsilonClosure(nfa->transition[0]);
	if (nfa->input[1] == EPSILON)
		if (AddMember(nfa->transition[1]))
			EpsilonClosure(nfa->transition[1]);
}


/* DFAを作る */
void MakeDfa()
{
	dfa_state_ptr Closure(closure_ptr, char);
	dfa_state_ptr cdp;
	char ch;

	for (cdp = start_of_dfa; cdp != NULL; cdp = cdp->next) {
		for (ch = ' '; ch < (char)CHAR_STAFF; ch++) {
			if (used[0x00ff & (int)ch]) {
				cdp->transition[0x00ff & (int)ch]
					= Closure(cdp->set_of_state, ch);
			}
		}
	}
}
		

/* 受理状態なら、構造体の状態記述部分に受理状態であるとマークする */
void MarkFinalState(dfa_state_ptr dfa)
{
	closure_ptr c;

	for ( ; dfa != NULL; dfa = dfa->next)
		for (c = dfa->set_of_state; c != NULL; c = c->next)
			if (c->state->input[0] == GOAL) {
				dfa->condition = GOAL;
				break;
			}
}


/* 文字cについての閉包を作る */
dfa_state_ptr Closure(closure_ptr c, char in)
{
	dfa_state_ptr MakeNewDfaState();
	dfa_state_ptr p;
	int flag = FALSE;

	for (cp = NULL; c != NULL; c = c->next) {
		if (c->state->input[0] == in) {
			flag = TRUE;
			AddMember(c->state->transition[0]);
			EpsilonClosure(c->state->transition[0]);
		}
	}

	if (!flag)
		return (NULL);

	for (p = start_of_dfa; p != NULL; p = p->next)
		if (EqualClosure(p->set_of_state, cp))
			return (p);

	p = MakeNewDfaState();
	p->set_of_state = cp;
	AddDfaState(p);
	return (p);
}


/* 新しいDFAの状態を、状態のチェインの最後尾に追加する */
void AddDfaState(dfa_state_ptr new)
{
	dfa_state_ptr p;

	for (p = start_of_dfa; p->next != NULL; p = p->next)
		;
	p->next = new;
}


/* 等価な閉包はいらない */
int EqualClosure(closure_ptr c1, closure_ptr c2)
{
	closure_ptr c;

	for (c = c2 ; c1 != NULL; c1 = c1->next, c = c->next)
		if (c1->state != c->state)
			return (FALSE);
	
	if (c == NULL) {
		FreeClosure(c2);
		return (TRUE);
	} else
		return (FALSE);
}


/* 閉包にメンバーを追加する */
int AddMember(nfa_state_ptr n)
{
	closure_ptr MakeNewClosure();
	closure_ptr p, tmp;

	if (cp == NULL) {
		cp = MakeNewClosure();
		cp->state = n;
		return (TRUE);
	}

	for (p = cp; p->next != NULL; p = p->next)
		if (p->state > n)
			break;
		else if (p->state == n)
			return (FALSE);

	if (p->state == n)
		return (FALSE);
	tmp = p->next;
	p->next = MakeNewClosure();
	(p->next)->next = tmp;
	if ((tmp == NULL) && (p->state < n))
		(p->next)->state = n;
	else {
		(p->next)->state = p->state;
		p->state = n;
	}
	return (TRUE);
}


/* 新しいDFAのためにメモリを確保する */
dfa_state_ptr MakeNewDfaState()
{
	dfa_state_ptr new;

	if ((new = (dfa_state_ptr)malloc(sizeof(dfa_state))) == NULL)
		FatalError("malloc",200);
	bzero(new, sizeof(dfa_state));
	//memset(new, 0, sizeof(dfa_state));
	return (new);
}


/* 新しい閉包のためにメモリを確保する */
closure_ptr MakeNewClosure()
{
	closure_ptr new;

	if ((new = (closure_ptr)malloc(sizeof(closure))) == NULL)
		FatalError("malloc",201);
	bzero(new, sizeof(closure));
	//memset(new, 0, sizeof(closure));
	return (new);
}


/* いらない閉包のための領域を解放する */
void FreeClosure(closure_ptr c)
{
	if (c->next != NULL)
		FreeClosure(c->next);
	free(c);
}
