/* $Header: output.cv  1.1  91/06/07 10:53:58  Exp $ */
/* output nfa & dfa tree for graph function */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "reg.h"

#define ABS(i) (i<0 ? -(i) : i)

typedef struct maplist *maplist_ptr;
typedef struct maplist {
	int		serial;
	long		source;
	maplist_ptr	next;
} maplist;

static maplist_ptr mlist = NULL;
static int max_node = 0;

static void ClearmapList(maplist_ptr);
static void NfaTree(nfa_state_ptr);

void FatalError(char *, int);

int map(long);

/* メモリを解放する */
static void ClearMapList(maplist_ptr ml)
{
	max_node = 0;
	if (ml == NULL)
		return;
	else {
		ClearMapList(ml->next);
		free(ml);
	}
}


/* NFAの木を表示する */
int OutputNfaTree(nfa_state_ptr nfa)
{
	maplist_ptr MakeNewMapList();

	ClearMapList(mlist);
	mlist = MakeNewMapList();
	NfaTree(nfa);
	return (max_node);
}

/* NFAの木を再帰的にたどって、全ての状態を出力する */
static void NfaTree(nfa_state_ptr nfa)
{
	int ii,i,f;

	for (ii=0; ii<2; ii++) {
		if (nfa->transition[ii] != NULL) {
			i = map((long)nfa);
			f = map((long)(nfa->transition[ii]));
			printf("% 02d\t% 02d\t",
				ABS(i), ABS(f) * (nfa->transition[ii]->input[0] == GOAL ? -1 : 1));
			switch (nfa->input[ii]) {
			 case EPSILON :
				printf("EPSILON\n");
				break;
			 default :
				printf("%c\n",nfa->input[ii]);
				break;
			}
			if (f>0)
				NfaTree(nfa->transition[ii]);
		}
	}
}


/* DFAの木を出力する */
int OutputDfaTree(dfa_state_ptr dfa)
{
	maplist_ptr MakeNewMapList();
	int ii;
	int i, f;

	ClearMapList(mlist);
	mlist = MakeNewMapList();

	for ( ; dfa != NULL; dfa = dfa->next) {
		for (ii=0x20 /* =' '*/; ii<CHAR_STAFF; ii++)
			if (dfa->transition[ii] != NULL) {
				i = map((long)dfa);
				f = map((long)dfa->transition[ii]);
				printf("% 02d\t% 02d\t%c\n",
					(ABS(i) * (dfa->condition ? -1 : 1)),
					(ABS(f) * (dfa->transition[ii]->condition ? -1 : 1)), (char)ii);
			}
	}
	return (max_node);
}


/* アドレスで示されたIDを、整理された通し番号に変換する */
int map(long l)
{
	maplist_ptr MakeNewMapList();
	maplist_ptr ml;
	int i;

	for (ml = mlist, i = 1; ml->next != NULL;
					ml = ml->next, i++) {
		if (ml->source == l)
			return (-ml->serial);
	}
	ml->source = l;
	ml->serial = i;
	ml->next = MakeNewMapList();
	max_node = i;
	return (ml->serial);
}


/* 変換のためのメモリを確保する */
maplist_ptr MakeNewMapList()
{
	maplist_ptr new;

	if ((new = (maplist_ptr)malloc(sizeof(maplist))) == NULL)
		FatalError("malloc",301);
	bzero(new, sizeof(maplist));
	//memset(new, 0, sizeof(maplist));
	return (new);
}
